"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const date_time_1 = __importDefault(require("date-time"));
const fs_1 = __importDefault(require("fs"));
const pretty_ms_1 = __importDefault(require("pretty-ms"));
const rollup = __importStar(require("rollup"));
const signal_exit_1 = __importDefault(require("signal-exit"));
const turbocolor_1 = __importDefault(require("turbocolor"));
const mergeOptions_1 = __importDefault(require("../../../src/utils/mergeOptions"));
const relativeId_1 = __importDefault(require("../../../src/utils/relativeId"));
const logging_1 = require("../logging");
const alternateScreen_1 = __importDefault(require("./alternateScreen"));
const batchWarnings_1 = __importDefault(require("./batchWarnings"));
const loadConfigFile_1 = __importDefault(require("./loadConfigFile"));
const timings_1 = require("./timings");
function watch(configFile, configs, command, silent = false) {
    const isTTY = Boolean(process.stderr.isTTY);
    const warnings = batchWarnings_1.default();
    const initialConfigs = processConfigs(configs);
    const clearScreen = initialConfigs.every(config => config.watch.clearScreen !== false);
    const screen = alternateScreen_1.default(isTTY && clearScreen);
    screen.open();
    let watcher;
    let configWatcher;
    let processConfigsErr;
    function processConfigs(configs) {
        return configs.map(options => {
            const merged = mergeOptions_1.default({
                command,
                config: options,
                defaultOnWarnHandler: warnings.add
            });
            const result = Object.assign(Object.assign({}, merged.inputOptions), { output: merged.outputOptions });
            if (!result.watch)
                result.watch = {};
            if (merged.optionError)
                merged.inputOptions.onwarn({
                    code: 'UNKNOWN_OPTION',
                    message: merged.optionError
                });
            if (merged.inputOptions.watch &&
                merged.inputOptions.watch.clearScreen === false) {
                processConfigsErr = logging_1.stderr;
            }
            return result;
        });
    }
    function start(configs) {
        const screenWriter = processConfigsErr || screen.reset;
        watcher = rollup.watch(configs);
        watcher.on('event', (event) => {
            switch (event.code) {
                case 'FATAL':
                    screen.close();
                    logging_1.handleError(event.error, true);
                    process.exit(1);
                    break;
                case 'ERROR':
                    warnings.flush();
                    logging_1.handleError(event.error, true);
                    break;
                case 'START':
                    if (!silent) {
                        screenWriter(turbocolor_1.default.underline(`rollup v${rollup.VERSION}`));
                    }
                    break;
                case 'BUNDLE_START':
                    if (!silent) {
                        let input = event.input;
                        if (typeof input !== 'string') {
                            input = Array.isArray(input)
                                ? input.join(', ')
                                : Object.keys(input)
                                    .map(key => input[key])
                                    .join(', ');
                        }
                        logging_1.stderr(turbocolor_1.default.cyan(`bundles ${turbocolor_1.default.bold(input)} → ${turbocolor_1.default.bold(event.output.map(relativeId_1.default).join(', '))}...`));
                    }
                    break;
                case 'BUNDLE_END':
                    warnings.flush();
                    if (!silent)
                        logging_1.stderr(turbocolor_1.default.green(`created ${turbocolor_1.default.bold(event.output.map(relativeId_1.default).join(', '))} in ${turbocolor_1.default.bold(pretty_ms_1.default(event.duration))}`));
                    if (event.result && event.result.getTimings) {
                        timings_1.printTimings(event.result.getTimings());
                    }
                    break;
                case 'END':
                    if (!silent && isTTY) {
                        logging_1.stderr(`\n[${date_time_1.default()}] waiting for changes...`);
                    }
            }
        });
    }
    // catch ctrl+c, kill, and uncaught errors
    const removeOnExit = signal_exit_1.default(close);
    process.on('uncaughtException', close);
    // only listen to stdin if it is a pipe
    if (!process.stdin.isTTY) {
        process.stdin.on('end', close); // in case we ever support stdin!
    }
    function close(err) {
        removeOnExit();
        process.removeListener('uncaughtException', close);
        // removing a non-existent listener is a no-op
        process.stdin.removeListener('end', close);
        screen.close();
        if (watcher)
            watcher.close();
        if (configWatcher)
            configWatcher.close();
        if (err) {
            console.error(err);
            process.exit(1);
        }
    }
    try {
        start(initialConfigs);
    }
    catch (err) {
        close(err);
        return;
    }
    if (configFile && !configFile.startsWith('node:')) {
        let restarting = false;
        let aborted = false;
        let configFileData = fs_1.default.readFileSync(configFile, 'utf-8');
        const restart = () => {
            const newConfigFileData = fs_1.default.readFileSync(configFile, 'utf-8');
            if (newConfigFileData === configFileData)
                return;
            configFileData = newConfigFileData;
            if (restarting) {
                aborted = true;
                return;
            }
            restarting = true;
            loadConfigFile_1.default(configFile, command)
                .then((_configs) => {
                restarting = false;
                if (aborted) {
                    aborted = false;
                    restart();
                }
                else {
                    watcher.close();
                    start(initialConfigs);
                }
            })
                .catch((err) => {
                logging_1.handleError(err, true);
            });
        };
        configWatcher = fs_1.default.watch(configFile, (event) => {
            if (event === 'change')
                restart();
        });
    }
}
exports.default = watch;
