"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const acorn = __importStar(require("acorn"));
const acorn_bigint_1 = __importDefault(require("acorn-bigint"));
const acorn_dynamic_import_1 = __importDefault(require("acorn-dynamic-import"));
const acorn_import_meta_1 = __importDefault(require("acorn-import-meta"));
const GlobalScope_1 = __importDefault(require("./ast/scopes/GlobalScope"));
const EntityPathTracker_1 = require("./ast/utils/EntityPathTracker");
const Chunk_1 = __importStar(require("./Chunk"));
const Module_1 = __importStar(require("./Module"));
const ModuleLoader_1 = require("./ModuleLoader");
const assetHooks_1 = require("./utils/assetHooks");
const buildPhase_1 = require("./utils/buildPhase");
const chunkColouring_1 = require("./utils/chunkColouring");
const entryHashing_1 = require("./utils/entryHashing");
const executionOrder_1 = require("./utils/executionOrder");
const path_1 = require("./utils/path");
const pluginDriver_1 = require("./utils/pluginDriver");
const relativeId_1 = __importDefault(require("./utils/relativeId"));
const timers_1 = require("./utils/timers");
function makeOnwarn() {
    const warned = Object.create(null);
    return (warning) => {
        const str = warning.toString();
        if (str in warned)
            return;
        console.error(str);
        warned[str] = true;
    };
}
function normalizeEntryModules(entryModules) {
    if (typeof entryModules === 'string') {
        return [{ alias: null, unresolvedId: entryModules }];
    }
    if (Array.isArray(entryModules)) {
        return entryModules.map(unresolvedId => ({ alias: null, unresolvedId }));
    }
    return Object.keys(entryModules).map(alias => ({
        alias,
        unresolvedId: entryModules[alias]
    }));
}
class Graph {
    constructor(options, watcher) {
        this.assetsById = new Map();
        this.curChunkIndex = 0;
        this.moduleById = new Map();
        this.needsTreeshakingPass = false;
        this.phase = buildPhase_1.BuildPhase.LOAD_AND_PARSE;
        this.watchFiles = Object.create(null);
        this.externalModules = [];
        this.modules = [];
        this.curChunkIndex = 0;
        this.deoptimizationTracker = new EntityPathTracker_1.EntityPathTracker();
        this.cachedModules = new Map();
        if (options.cache) {
            if (options.cache.modules)
                for (const module of options.cache.modules)
                    this.cachedModules.set(module.id, module);
        }
        if (options.cache !== false) {
            this.pluginCache = (options.cache && options.cache.plugins) || Object.create(null);
            // increment access counter
            for (const name in this.pluginCache) {
                const cache = this.pluginCache[name];
                for (const key of Object.keys(cache))
                    cache[key][0]++;
            }
        }
        this.preserveModules = options.preserveModules;
        this.cacheExpiry = options.experimentalCacheExpiry;
        this.treeshake = options.treeshake !== false;
        if (this.treeshake) {
            this.treeshakingOptions = options.treeshake
                ? {
                    annotations: options.treeshake.annotations !== false,
                    moduleSideEffects: options.treeshake.moduleSideEffects,
                    propertyReadSideEffects: options.treeshake.propertyReadSideEffects !== false,
                    pureExternalModules: options.treeshake.pureExternalModules
                }
                : {
                    annotations: true,
                    moduleSideEffects: true,
                    propertyReadSideEffects: true,
                    pureExternalModules: false
                };
        }
        this.contextParse = (code, options = {}) => this.acornParser.parse(code, Object.assign(Object.assign(Object.assign({}, Module_1.defaultAcornOptions), options), this.acornOptions));
        this.pluginDriver = pluginDriver_1.createPluginDriver(this, options, this.pluginCache, watcher);
        if (watcher) {
            const handleChange = (id) => this.pluginDriver.hookSeqSync('watchChange', [id]);
            watcher.on('change', handleChange);
            watcher.once('restart', () => {
                watcher.removeListener('change', handleChange);
            });
        }
        this.shimMissingExports = options.shimMissingExports;
        this.scope = new GlobalScope_1.default();
        this.context = String(options.context);
        const optionsModuleContext = options.moduleContext;
        if (typeof optionsModuleContext === 'function') {
            this.getModuleContext = id => optionsModuleContext(id) || this.context;
        }
        else if (typeof optionsModuleContext === 'object') {
            const moduleContext = new Map();
            for (const key in optionsModuleContext) {
                moduleContext.set(path_1.resolve(key), optionsModuleContext[key]);
            }
            this.getModuleContext = id => moduleContext.get(id) || this.context;
        }
        else {
            this.getModuleContext = () => this.context;
        }
        this.onwarn = options.onwarn || makeOnwarn();
        this.acornOptions = options.acorn || {};
        const acornPluginsToInject = [];
        acornPluginsToInject.push(acorn_dynamic_import_1.default);
        acornPluginsToInject.push(acorn_import_meta_1.default);
        acornPluginsToInject.push(acorn_bigint_1.default);
        if (options.experimentalTopLevelAwait) {
            this.acornOptions.allowAwaitOutsideFunction = true;
        }
        const acornInjectPlugins = options.acornInjectPlugins;
        acornPluginsToInject.push(...(Array.isArray(acornInjectPlugins)
            ? acornInjectPlugins
            : acornInjectPlugins
                ? [acornInjectPlugins]
                : []));
        this.acornParser = acorn.Parser.extend(...acornPluginsToInject);
        this.moduleLoader = new ModuleLoader_1.ModuleLoader(this, this.moduleById, this.pluginDriver, options.external, typeof options.manualChunks === 'function' && options.manualChunks, this.treeshake ? this.treeshakingOptions.moduleSideEffects : null, this.treeshake ? this.treeshakingOptions.pureExternalModules : false);
    }
    build(entryModules, manualChunks, inlineDynamicImports) {
        // Phase 1 – discovery. We load the entry module and find which
        // modules it imports, and import those, until we have all
        // of the entry module's dependencies
        timers_1.timeStart('parse modules', 2);
        return Promise.all([
            this.moduleLoader.addEntryModules(normalizeEntryModules(entryModules), true),
            manualChunks &&
                typeof manualChunks === 'object' &&
                this.moduleLoader.addManualChunks(manualChunks)
        ]).then(([{ entryModules, manualChunkModulesByAlias }]) => {
            if (entryModules.length === 0) {
                throw new Error('You must supply options.input to rollup');
            }
            for (const module of this.moduleById.values()) {
                if (module instanceof Module_1.default) {
                    this.modules.push(module);
                    this.watchFiles[module.id] = true;
                }
                else {
                    this.externalModules.push(module);
                }
            }
            timers_1.timeEnd('parse modules', 2);
            this.phase = buildPhase_1.BuildPhase.ANALYSE;
            // Phase 2 - linking. We populate the module dependency links and
            // determine the topological execution order for the bundle
            timers_1.timeStart('analyse dependency graph', 2);
            this.link(entryModules);
            timers_1.timeEnd('analyse dependency graph', 2);
            // Phase 3 – marking. We include all statements that should be included
            timers_1.timeStart('mark included statements', 2);
            if (inlineDynamicImports) {
                if (entryModules.length > 1) {
                    throw new Error('Internal Error: can only inline dynamic imports for single-file builds.');
                }
            }
            for (const module of entryModules) {
                module.includeAllExports();
            }
            this.includeMarked(this.modules);
            // check for unused external imports
            for (const externalModule of this.externalModules)
                externalModule.warnUnusedImports();
            timers_1.timeEnd('mark included statements', 2);
            // Phase 4 – we construct the chunks, working out the optimal chunking using
            // entry point graph colouring, before generating the import and export facades
            timers_1.timeStart('generate chunks', 2);
            if (!this.preserveModules && !inlineDynamicImports) {
                chunkColouring_1.assignChunkColouringHashes(entryModules, manualChunkModulesByAlias);
            }
            // TODO: there is one special edge case unhandled here and that is that any module
            //       exposed as an unresolvable export * (to a graph external export *,
            //       either as a namespace import reexported or top-level export *)
            //       should be made to be its own entry point module before chunking
            let chunks = [];
            if (this.preserveModules) {
                for (const module of this.modules) {
                    const chunk = new Chunk_1.default(this, [module]);
                    if (module.isEntryPoint || !chunk.isEmpty) {
                        chunk.entryModules = [module];
                    }
                    chunks.push(chunk);
                }
            }
            else {
                const chunkModules = {};
                for (const module of this.modules) {
                    const entryPointsHashStr = entryHashing_1.Uint8ArrayToHexString(module.entryPointsHash);
                    const curChunk = chunkModules[entryPointsHashStr];
                    if (curChunk) {
                        curChunk.push(module);
                    }
                    else {
                        chunkModules[entryPointsHashStr] = [module];
                    }
                }
                for (const entryHashSum in chunkModules) {
                    const chunkModulesOrdered = chunkModules[entryHashSum];
                    executionOrder_1.sortByExecutionOrder(chunkModulesOrdered);
                    const chunk = new Chunk_1.default(this, chunkModulesOrdered);
                    chunks.push(chunk);
                }
            }
            // for each chunk module, set up its imports to other
            // chunks, if those variables are included after treeshaking
            for (const chunk of chunks) {
                chunk.link();
            }
            // filter out empty dependencies
            chunks = chunks.filter(Chunk_1.isChunkRendered);
            // then go over and ensure all entry chunks export their variables
            for (const chunk of chunks) {
                if (this.preserveModules || chunk.entryModules.length > 0) {
                    chunk.generateEntryExportsOrMarkAsTainted();
                }
            }
            // create entry point facades for entry module chunks that have tainted exports
            const facades = [];
            if (!this.preserveModules) {
                for (const chunk of chunks) {
                    for (const entryModule of chunk.entryModules) {
                        if (chunk.facadeModule !== entryModule) {
                            const entryPointFacade = new Chunk_1.default(this, []);
                            entryPointFacade.turnIntoFacade(entryModule);
                            facades.push(entryPointFacade);
                        }
                    }
                }
            }
            timers_1.timeEnd('generate chunks', 2);
            this.phase = buildPhase_1.BuildPhase.GENERATE;
            return chunks.concat(facades);
        });
    }
    finaliseAssets(assetFileNames) {
        const outputBundle = Object.create(null);
        this.assetsById.forEach(asset => {
            if (asset.source !== undefined)
                assetHooks_1.finaliseAsset(asset, outputBundle, assetFileNames);
        });
        return outputBundle;
    }
    getCache() {
        // handle plugin cache eviction
        for (const name in this.pluginCache) {
            const cache = this.pluginCache[name];
            let allDeleted = true;
            for (const key of Object.keys(cache)) {
                if (cache[key][0] >= this.cacheExpiry)
                    delete cache[key];
                else
                    allDeleted = false;
            }
            if (allDeleted)
                delete this.pluginCache[name];
        }
        return {
            modules: this.modules.map(module => module.toJSON()),
            plugins: this.pluginCache
        };
    }
    includeMarked(modules) {
        if (this.treeshake) {
            let treeshakingPass = 1;
            do {
                timers_1.timeStart(`treeshaking pass ${treeshakingPass}`, 3);
                this.needsTreeshakingPass = false;
                for (const module of modules) {
                    if (module.isExecuted)
                        module.include();
                }
                timers_1.timeEnd(`treeshaking pass ${treeshakingPass++}`, 3);
            } while (this.needsTreeshakingPass);
        }
        else {
            // Necessary to properly replace namespace imports
            for (const module of modules)
                module.includeAllInBundle();
        }
    }
    warn(warning) {
        warning.toString = () => {
            let str = '';
            if (warning.plugin)
                str += `(${warning.plugin} plugin) `;
            if (warning.loc)
                str += `${relativeId_1.default(warning.loc.file)} (${warning.loc.line}:${warning.loc.column}) `;
            str += warning.message;
            return str;
        };
        this.onwarn(warning);
    }
    link(entryModules) {
        for (const module of this.modules) {
            module.linkDependencies();
        }
        const { orderedModules, cyclePaths } = executionOrder_1.analyseModuleExecution(entryModules);
        for (const cyclePath of cyclePaths) {
            this.warn({
                code: 'CIRCULAR_DEPENDENCY',
                importer: cyclePath[0],
                message: `Circular dependency: ${cyclePath.join(' -> ')}`
            });
        }
        this.modules = orderedModules;
        for (const module of this.modules) {
            module.bindReferences();
        }
        this.warnForMissingExports();
    }
    warnForMissingExports() {
        for (const module of this.modules) {
            for (const importName of Object.keys(module.importDescriptions)) {
                const importDescription = module.importDescriptions[importName];
                if (importDescription.name !== '*' &&
                    !importDescription.module.getVariableForExportName(importDescription.name)) {
                    module.warn({
                        code: 'NON_EXISTENT_EXPORT',
                        message: `Non-existent export '${importDescription.name}' is imported from ${relativeId_1.default(importDescription.module.id)}`,
                        name: importDescription.name,
                        source: importDescription.module.id
                    }, importDescription.start);
                }
            }
        }
    }
}
exports.default = Graph;
