"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ExternalModule_1 = __importDefault(require("./ExternalModule"));
const Module_1 = __importDefault(require("./Module"));
const error_1 = require("./utils/error");
const path_1 = require("./utils/path");
const referenceIds_1 = require("./utils/referenceIds");
const relativeId_1 = __importDefault(require("./utils/relativeId"));
const timers_1 = require("./utils/timers");
const transform_1 = __importDefault(require("./utils/transform"));
function normalizeRelativeExternalId(importer, source) {
    return path_1.isRelative(source) ? path_1.resolve(importer, '..', source) : source;
}
function getIdMatcher(option) {
    if (option === true) {
        return () => true;
    }
    else if (typeof option === 'function') {
        return (id, ...args) => (!id.startsWith('\0') && option(id, ...args)) || false;
    }
    else if (option) {
        const ids = new Set(Array.isArray(option) ? option : option ? [option] : []);
        return id => ids.has(id);
    }
    else {
        return () => false;
    }
}
function getHasModuleSideEffects(moduleSideEffectsOption, pureExternalModules, graph) {
    if (typeof moduleSideEffectsOption === 'boolean') {
        return () => moduleSideEffectsOption;
    }
    if (moduleSideEffectsOption === 'no-external') {
        return (_id, external) => !external;
    }
    if (typeof moduleSideEffectsOption === 'function') {
        return (id, external) => !id.startsWith('\0') ? moduleSideEffectsOption(id, external) !== false : true;
    }
    if (Array.isArray(moduleSideEffectsOption)) {
        const ids = new Set(moduleSideEffectsOption);
        return id => ids.has(id);
    }
    if (moduleSideEffectsOption) {
        graph.warn(error_1.errInvalidOption('treeshake.moduleSideEffects', 'please use one of false, "no-external", a function or an array'));
    }
    const isPureExternalModule = getIdMatcher(pureExternalModules);
    return (id, external) => !(external && isPureExternalModule(id));
}
class ModuleLoader {
    constructor(graph, modulesById, pluginDriver, external, getManualChunk, moduleSideEffects, pureExternalModules) {
        this.entriesByReferenceId = new Map();
        this.entryModules = [];
        this.latestLoadModulesPromise = Promise.resolve();
        this.manualChunkModules = {};
        this.loadEntryModule = ({ alias, unresolvedId }, isEntry) => this.pluginDriver
            .hookFirst('resolveId', [unresolvedId, undefined])
            .then((resolveIdResult) => {
            if (resolveIdResult === false ||
                (resolveIdResult && typeof resolveIdResult === 'object' && resolveIdResult.external)) {
                error_1.error(error_1.errEntryCannotBeExternal(unresolvedId));
            }
            const id = resolveIdResult && typeof resolveIdResult === 'object'
                ? resolveIdResult.id
                : resolveIdResult;
            if (typeof id === 'string') {
                return this.fetchModule(id, undefined, true, isEntry).then(module => {
                    if (alias !== null) {
                        if (module.chunkAlias !== null && module.chunkAlias !== alias) {
                            error_1.error(error_1.errCannotAssignModuleToChunk(module.id, alias, module.chunkAlias));
                        }
                        module.chunkAlias = alias;
                    }
                    return module;
                });
            }
            error_1.error(error_1.errUnresolvedEntry(unresolvedId));
        });
        this.graph = graph;
        this.modulesById = modulesById;
        this.pluginDriver = pluginDriver;
        this.isExternal = getIdMatcher(external);
        this.hasModuleSideEffects = getHasModuleSideEffects(moduleSideEffects, pureExternalModules, graph);
        this.getManualChunk = typeof getManualChunk === 'function' ? getManualChunk : () => null;
    }
    addEntryModuleAndGetReferenceId(unresolvedEntryModule) {
        const entryRecord = {
            module: null,
            name: unresolvedEntryModule.unresolvedId
        };
        const referenceId = referenceIds_1.addWithNewReferenceId(entryRecord, this.entriesByReferenceId, unresolvedEntryModule.unresolvedId);
        this.addEntryModules([unresolvedEntryModule], false)
            .then(({ newEntryModules: [module] }) => {
            entryRecord.module = module;
        })
            .catch(() => {
            // Avoid unhandled Promise rejection as the error will be thrown later
            // once module loading has finished
        });
        return referenceId;
    }
    addEntryModules(unresolvedEntryModules, isUserDefined) {
        const loadNewEntryModulesPromise = Promise.all(unresolvedEntryModules.map(unresolvedEntryModule => this.loadEntryModule(unresolvedEntryModule, true))).then(entryModules => {
            for (const entryModule of entryModules) {
                entryModule.isUserDefinedEntryPoint = entryModule.isUserDefinedEntryPoint || isUserDefined;
                const existingEntryModule = this.entryModules.find(module => module.id === entryModule.id);
                if (!existingEntryModule) {
                    this.entryModules.push(entryModule);
                }
            }
            return entryModules;
        });
        return this.awaitLoadModulesPromise(loadNewEntryModulesPromise).then(newEntryModules => ({
            entryModules: this.entryModules,
            manualChunkModulesByAlias: this.manualChunkModules,
            newEntryModules
        }));
    }
    addManualChunks(manualChunks) {
        const unresolvedManualChunks = [];
        for (const alias of Object.keys(manualChunks)) {
            const manualChunkIds = manualChunks[alias];
            for (const unresolvedId of manualChunkIds) {
                unresolvedManualChunks.push({ alias: null, unresolvedId, manualChunkAlias: alias });
            }
        }
        const loadNewManualChunkModulesPromise = Promise.all(unresolvedManualChunks.map(unresolvedManualChunk => this.loadEntryModule(unresolvedManualChunk, false))).then(manualChunkModules => {
            for (let index = 0; index < manualChunkModules.length; index++) {
                this.addToManualChunk(unresolvedManualChunks[index].manualChunkAlias, manualChunkModules[index]);
            }
        });
        return this.awaitLoadModulesPromise(loadNewManualChunkModulesPromise);
    }
    getChunkFileName(referenceId) {
        const entryRecord = this.entriesByReferenceId.get(referenceId);
        if (!entryRecord)
            error_1.error(error_1.errChunkReferenceIdNotFoundForFilename(referenceId));
        const fileName = entryRecord.module &&
            (entryRecord.module.facadeChunk
                ? entryRecord.module.facadeChunk.id
                : entryRecord.module.chunk.id);
        if (!fileName)
            error_1.error(error_1.errChunkNotGeneratedForFileName(entryRecord));
        return fileName;
    }
    resolveId(source, importer, skip) {
        return Promise.resolve(this.isExternal(source, importer, false)
            ? { id: source, external: true }
            : this.pluginDriver.hookFirst('resolveId', [source, importer], null, skip)).then((result) => this.normalizeResolveIdResult(result, importer, source));
    }
    addToManualChunk(alias, module) {
        if (module.manualChunkAlias !== null && module.manualChunkAlias !== alias) {
            error_1.error(error_1.errCannotAssignModuleToChunk(module.id, alias, module.manualChunkAlias));
        }
        module.manualChunkAlias = alias;
        if (!this.manualChunkModules[alias]) {
            this.manualChunkModules[alias] = [];
        }
        this.manualChunkModules[alias].push(module);
    }
    awaitLoadModulesPromise(loadNewModulesPromise) {
        this.latestLoadModulesPromise = Promise.all([
            loadNewModulesPromise,
            this.latestLoadModulesPromise
        ]);
        const getCombinedPromise = () => {
            const startingPromise = this.latestLoadModulesPromise;
            return startingPromise.then(() => {
                if (this.latestLoadModulesPromise !== startingPromise) {
                    return getCombinedPromise();
                }
            });
        };
        return getCombinedPromise().then(() => loadNewModulesPromise);
    }
    fetchAllDependencies(module) {
        const fetchDynamicImportsPromise = Promise.all(module.getDynamicImportExpressions().map((specifier, index) => this.resolveDynamicImport(specifier, module.id).then(resolvedId => {
            if (resolvedId === null)
                return;
            const dynamicImport = module.dynamicImports[index];
            if (typeof resolvedId === 'string') {
                dynamicImport.resolution = resolvedId;
                return;
            }
            return this.fetchResolvedDependency(relativeId_1.default(resolvedId.id), module.id, resolvedId).then(module => {
                dynamicImport.resolution = module;
            });
        })));
        fetchDynamicImportsPromise.catch(() => { });
        return Promise.all(module.sources.map(source => this.resolveAndFetchDependency(module, source))).then(() => fetchDynamicImportsPromise);
    }
    fetchModule(id, importer, moduleSideEffects, isEntry) {
        const existingModule = this.modulesById.get(id);
        if (existingModule) {
            if (existingModule instanceof ExternalModule_1.default)
                throw new Error(`Cannot fetch external module ${id}`);
            existingModule.isEntryPoint = existingModule.isEntryPoint || isEntry;
            return Promise.resolve(existingModule);
        }
        const module = new Module_1.default(this.graph, id, moduleSideEffects, isEntry);
        this.modulesById.set(id, module);
        const manualChunkAlias = this.getManualChunk(id);
        if (typeof manualChunkAlias === 'string') {
            this.addToManualChunk(manualChunkAlias, module);
        }
        timers_1.timeStart('load modules', 3);
        return Promise.resolve(this.pluginDriver.hookFirst('load', [id]))
            .catch((err) => {
            timers_1.timeEnd('load modules', 3);
            let msg = `Could not load ${id}`;
            if (importer)
                msg += ` (imported by ${importer})`;
            msg += `: ${err.message}`;
            throw new Error(msg);
        })
            .then(source => {
            timers_1.timeEnd('load modules', 3);
            if (typeof source === 'string')
                return { code: source };
            if (source && typeof source === 'object' && typeof source.code === 'string')
                return source;
            error_1.error(error_1.errBadLoader(id));
        })
            .then(sourceDescription => {
            const cachedModule = this.graph.cachedModules.get(id);
            if (cachedModule &&
                !cachedModule.customTransformCache &&
                cachedModule.originalCode === sourceDescription.code) {
                // re-emit transform assets
                if (cachedModule.transformAssets) {
                    for (const asset of cachedModule.transformAssets)
                        this.pluginDriver.emitAsset(asset.name, asset.source);
                }
                return cachedModule;
            }
            if (typeof sourceDescription.moduleSideEffects === 'boolean') {
                module.moduleSideEffects = sourceDescription.moduleSideEffects;
            }
            return transform_1.default(this.graph, sourceDescription, module);
        })
            .then((source) => {
            module.setSource(source);
            this.modulesById.set(id, module);
            return this.fetchAllDependencies(module).then(() => {
                for (const name in module.exports) {
                    if (name !== 'default') {
                        module.exportsAll[name] = module.id;
                    }
                }
                module.exportAllSources.forEach(source => {
                    const id = module.resolvedIds[source].id;
                    const exportAllModule = this.modulesById.get(id);
                    if (exportAllModule instanceof ExternalModule_1.default)
                        return;
                    for (const name in exportAllModule.exportsAll) {
                        if (name in module.exportsAll) {
                            this.graph.warn(error_1.errNamespaceConflict(name, module, exportAllModule));
                        }
                        else {
                            module.exportsAll[name] = exportAllModule.exportsAll[name];
                        }
                    }
                });
                return module;
            });
        });
    }
    fetchResolvedDependency(source, importer, resolvedId) {
        if (resolvedId.external) {
            if (!this.modulesById.has(resolvedId.id)) {
                this.modulesById.set(resolvedId.id, new ExternalModule_1.default(this.graph, resolvedId.id, resolvedId.moduleSideEffects));
            }
            const externalModule = this.modulesById.get(resolvedId.id);
            if (externalModule instanceof ExternalModule_1.default === false) {
                error_1.error(error_1.errInternalIdCannotBeExternal(source, importer));
            }
            return Promise.resolve(externalModule);
        }
        else {
            return this.fetchModule(resolvedId.id, importer, resolvedId.moduleSideEffects, false);
        }
    }
    handleMissingImports(resolvedId, source, importer) {
        if (resolvedId === null) {
            if (path_1.isRelative(source)) {
                error_1.error(error_1.errUnresolvedImport(source, importer));
            }
            this.graph.warn(error_1.errUnresolvedImportTreatedAsExternal(source, importer));
            return { id: source, external: true, moduleSideEffects: true };
        }
        return resolvedId;
    }
    normalizeResolveIdResult(resolveIdResult, importer, source) {
        let id = '';
        let external = false;
        let moduleSideEffects = null;
        if (resolveIdResult) {
            if (typeof resolveIdResult === 'object') {
                id = resolveIdResult.id;
                if (resolveIdResult.external) {
                    external = true;
                }
                if (typeof resolveIdResult.moduleSideEffects === 'boolean') {
                    moduleSideEffects = resolveIdResult.moduleSideEffects;
                }
            }
            else {
                id = resolveIdResult;
                if (this.isExternal(id, importer, true)) {
                    external = true;
                }
            }
            if (external) {
                id = normalizeRelativeExternalId(importer, id);
            }
        }
        else {
            id = normalizeRelativeExternalId(importer, source);
            if (resolveIdResult !== false && !this.isExternal(id, importer, true)) {
                return null;
            }
            external = true;
        }
        return {
            external,
            id,
            moduleSideEffects: typeof moduleSideEffects === 'boolean'
                ? moduleSideEffects
                : this.hasModuleSideEffects(id, external)
        };
    }
    resolveAndFetchDependency(module, source) {
        return Promise.resolve(module.resolvedIds[source] ||
            this.resolveId(source, module.id).then(resolvedId => this.handleMissingImports(resolvedId, source, module.id))).then(resolvedId => {
            module.resolvedIds[source] = resolvedId;
            return this.fetchResolvedDependency(source, module.id, resolvedId);
        });
    }
    resolveDynamicImport(specifier, importer) {
        // TODO we only should expose the acorn AST here
        return this.pluginDriver
            .hookFirst('resolveDynamicImport', [specifier, importer])
            .then((resolution) => {
            if (typeof specifier !== 'string') {
                if (typeof resolution === 'string') {
                    return resolution;
                }
                if (!resolution) {
                    return null;
                }
                return Object.assign({ external: false, moduleSideEffects: true }, resolution);
            }
            if (resolution == null) {
                return this.resolveId(specifier, importer).then(resolvedId => this.handleMissingImports(resolvedId, specifier, importer));
            }
            return this.handleMissingImports(this.normalizeResolveIdResult(resolution, importer, specifier), specifier, importer);
        });
    }
}
exports.ModuleLoader = ModuleLoader;
