"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const events_1 = require("events");
const path_1 = __importDefault(require("path"));
const rollup_pluginutils_1 = require("rollup-pluginutils");
const index_1 = __importStar(require("../rollup/index"));
const mergeOptions_1 = __importDefault(require("../utils/mergeOptions"));
const chokidar_1 = __importDefault(require("./chokidar"));
const fileWatchers_1 = require("./fileWatchers");
const DELAY = 200;
class Watcher {
    constructor(configs) {
        this.invalidatedIds = new Set();
        this.rerun = false;
        this.succeeded = false;
        this.emitter = new (class extends events_1.EventEmitter {
            constructor(close) {
                super();
                this.close = close;
                // Allows more than 10 bundles to be watched without
                // showing the `MaxListenersExceededWarning` to the user.
                this.setMaxListeners(Infinity);
            }
        })(this.close.bind(this));
        this.tasks = (Array.isArray(configs) ? configs : configs ? [configs] : []).map(config => new Task(this, config));
        this.running = true;
        process.nextTick(() => this.run());
    }
    close() {
        if (this.buildTimeout)
            clearTimeout(this.buildTimeout);
        this.tasks.forEach(task => {
            task.close();
        });
        this.emitter.removeAllListeners();
    }
    emit(event, value) {
        this.emitter.emit(event, value);
    }
    invalidate(id) {
        if (id) {
            this.invalidatedIds.add(id);
        }
        if (this.running) {
            this.rerun = true;
            return;
        }
        if (this.buildTimeout)
            clearTimeout(this.buildTimeout);
        this.buildTimeout = setTimeout(() => {
            this.buildTimeout = undefined;
            this.invalidatedIds.forEach(id => this.emit('change', id));
            this.invalidatedIds.clear();
            this.emit('restart');
            this.run();
        }, DELAY);
    }
    run() {
        this.running = true;
        this.emit('event', {
            code: 'START'
        });
        let taskPromise = Promise.resolve();
        for (const task of this.tasks)
            taskPromise = taskPromise.then(() => task.run());
        return taskPromise
            .then(() => {
            this.succeeded = true;
            this.running = false;
            this.emit('event', {
                code: 'END'
            });
        })
            .catch(error => {
            this.running = false;
            this.emit('event', {
                code: this.succeeded ? 'ERROR' : 'FATAL',
                error
            });
        })
            .then(() => {
            if (this.rerun) {
                this.rerun = false;
                this.invalidate();
            }
        });
    }
}
exports.Watcher = Watcher;
class Task {
    constructor(watcher, config) {
        this.invalidated = true;
        this.cache = null;
        this.watcher = watcher;
        this.closed = false;
        this.watched = new Set();
        const { inputOptions, outputOptions } = mergeOptions_1.default({
            config
        });
        this.inputOptions = inputOptions;
        this.outputs = outputOptions;
        this.outputFiles = this.outputs.map(output => {
            if (output.file || output.dir)
                return path_1.default.resolve(output.file || output.dir);
        });
        const watchOptions = inputOptions.watch || {};
        if ('useChokidar' in watchOptions)
            watchOptions.chokidar = watchOptions.useChokidar;
        let chokidarOptions = 'chokidar' in watchOptions ? watchOptions.chokidar : !!chokidar_1.default;
        if (chokidarOptions) {
            chokidarOptions = Object.assign(Object.assign({}, (chokidarOptions === true ? {} : chokidarOptions)), { disableGlobbing: true, ignoreInitial: true });
        }
        if (chokidarOptions && !chokidar_1.default) {
            throw new Error(`watch.chokidar was provided, but chokidar could not be found. Have you installed it?`);
        }
        this.chokidarOptions = chokidarOptions;
        this.chokidarOptionsHash = JSON.stringify(chokidarOptions);
        this.filter = rollup_pluginutils_1.createFilter(watchOptions.include, watchOptions.exclude);
    }
    close() {
        this.closed = true;
        this.watched.forEach(id => {
            fileWatchers_1.deleteTask(id, this, this.chokidarOptionsHash);
        });
    }
    invalidate(id, isTransformDependency) {
        this.invalidated = true;
        if (isTransformDependency) {
            this.cache.modules.forEach(module => {
                if (!module.transformDependencies || module.transformDependencies.indexOf(id) === -1)
                    return;
                // effective invalidation
                module.originalCode = null;
            });
        }
        this.watcher.invalidate(id);
    }
    run() {
        if (!this.invalidated)
            return;
        this.invalidated = false;
        const options = Object.assign(Object.assign({}, this.inputOptions), { cache: this.cache });
        const start = Date.now();
        this.watcher.emit('event', {
            code: 'BUNDLE_START',
            input: this.inputOptions.input,
            output: this.outputFiles
        });
        index_1.setWatcher(this.watcher.emitter);
        return index_1.default(options)
            .then(result => {
            if (this.closed)
                return;
            const watched = (this.watched = new Set());
            this.cache = result.cache;
            this.watchFiles = result.watchFiles;
            this.cache.modules.forEach(module => {
                if (module.transformDependencies) {
                    module.transformDependencies.forEach(depId => {
                        watched.add(depId);
                        this.watchFile(depId, true);
                    });
                }
            });
            this.watchFiles.forEach(id => {
                watched.add(id);
                this.watchFile(id);
            });
            this.watched.forEach(id => {
                if (!watched.has(id))
                    fileWatchers_1.deleteTask(id, this, this.chokidarOptionsHash);
            });
            return Promise.all(this.outputs.map(output => result.write(output))).then(() => result);
        })
            .then((result) => {
            this.watcher.emit('event', {
                code: 'BUNDLE_END',
                duration: Date.now() - start,
                input: this.inputOptions.input,
                output: this.outputFiles,
                result
            });
        })
            .catch((error) => {
            if (this.closed)
                return;
            if (this.cache) {
                // this is necessary to ensure that any 'renamed' files
                // continue to be watched following an error
                if (this.cache.modules) {
                    this.cache.modules.forEach(module => {
                        if (module.transformDependencies) {
                            module.transformDependencies.forEach(depId => {
                                this.watchFile(depId, true);
                            });
                        }
                    });
                }
                this.watchFiles.forEach(id => {
                    this.watchFile(id);
                });
            }
            throw error;
        });
    }
    watchFile(id, isTransformDependency = false) {
        if (!this.filter(id))
            return;
        if (this.outputFiles.some(file => file === id)) {
            throw new Error('Cannot import the generated bundle');
        }
        // this is necessary to ensure that any 'renamed' files
        // continue to be watched following an error
        fileWatchers_1.addTask(id, this, this.chokidarOptions, this.chokidarOptionsHash, isTransformDependency);
    }
}
exports.Task = Task;
function watch(configs) {
    return new Watcher(configs).emitter;
}
exports.default = watch;
